// Mayce wrapper
// devilducky@gmx.net

#include "tools.h"
#include "winfunc.h"
#include "dxdraw.h"
#include "ddutil.h"

MImage Screen;
MImage ScreenBack;

////////////////////////////
// MImage class functions //
////////////////////////////

void MImage::Create()
{
	Surface = NULL;
	Height = 0;
	Width = 0;
}
void MImage::Destroy()
{
	if (Surface) Surface->Release();
	Surface = NULL;
}
MImage::MImage() 
{
	Create();
}
MImage::MImage(LPCSTR strFile) 
{
	Create ();
	Load (strFile);
}
MImage::MImage (int iNewWidth, int iNewHeight)
{
	Create ();
	New (iNewWidth, iNewHeight);
}
MImage::~MImage() 
{
	Destroy();
}

MImage& MImage::operator =(MImage& mimOp)
{
	RECT rcAll = { 0, 0, Width, Height };
		
	if( !Surface || mimOp.Width != Width || mimOp.Height != Height )
	{
		if (!mimOp.Surface)
		{
			Destroy ();
			return *this;
		}

		// Create a new surface
		New (mimOp.Width, mimOp.Height);
	}
	CheckDD( Surface->BltFast( 0, 0, mimOp.Surface, &rcAll, DDBLTFAST_NOCOLORKEY | DDBLTFAST_WAIT ), "bltfast");
		
	return *this;
}
void MImage::Load (LPCSTR strFile) 
{
	DDSURFACEDESC2 dscOp;
	
	Destroy();
	Surface = DDLoadBitmap( Draw->lpDD, strFile, 0, 0);
	
	dscOp.dwSize = sizeof(DDSURFACEDESC2);
	Surface->GetSurfaceDesc (&dscOp);
	Width = dscOp.dwWidth;
	Height = dscOp.dwHeight;

	// set colorkey
	DDCOLORKEY ColorKey;
	ColorKey.dwColorSpaceLowValue = MCOLORKEY;
	ColorKey.dwColorSpaceHighValue = ColorKey.dwColorSpaceLowValue;

	CheckDD( Surface->SetColorKey (DDCKEY_SRCBLT, &ColorKey), "setcolorkey");
}
void MImage::Blit (int posX, int posY, MImage &mimSrc, int posXSrc, int posYSrc, int lenX, int lenY)
{
	RECT rcAll;
	rcAll.left = posXSrc;
	rcAll.top = posYSrc;
	rcAll.right = lenX + posXSrc;
	rcAll.bottom = lenY + posYSrc;
	
	CheckDD( Surface->BltFast( posX, posY, mimSrc.Surface, &rcAll, DDBLTFAST_NOCOLORKEY | DDBLTFAST_WAIT), "Bltfast");
}
void MImage::BlitTransparent (int posX, int posY, MImage &mimSrc)
{
	RECT rcAll;
	rcAll.left = 0;
	rcAll.top = 0;
	rcAll.right = mimSrc.Width;
	rcAll.bottom = mimSrc.Height;

	CheckDD( Surface->BltFast( posX, posY, mimSrc.Surface, &rcAll, DDBLTFAST_SRCCOLORKEY | DDBLTFAST_WAIT), "Bltfast");
}
void MImage::New (int iNewWidth, int iNewHeight)
{
	DDSURFACEDESC2 dscOp;

	Destroy ();

	Width = iNewWidth;
	Height = iNewHeight;
	memset (&dscOp, 0, sizeof (DDSURFACEDESC2));
	dscOp.dwSize = sizeof(DDSURFACEDESC2);
	dscOp.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	dscOp.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	dscOp.dwWidth = Width;
    dscOp.dwHeight = Height;
    
	CheckDD( Draw->lpDD->CreateSurface (&dscOp, &Surface, NULL), "create surface");
	
	// set colorkey
	DDCOLORKEY ColorKey;
	ColorKey.dwColorSpaceLowValue = MCOLORKEY;
	ColorKey.dwColorSpaceHighValue = ColorKey.dwColorSpaceLowValue;

	CheckDD( Surface->SetColorKey (DDCKEY_SRCBLT, &ColorKey), "setcolorkey");
}
COLORREF MImage::GetPoint (int posX, int posY)
{
	HDC hdcPix;
	CheckDD (Surface->GetDC (&hdcPix), "Pixel GetDC");

	COLORREF color = GetPixel (hdcPix, posX, posY);
	
	CheckDD (Surface->ReleaseDC (hdcPix), "Pixel releaseDC");
	return color;
}
void MImage::Clear (int posX0, int posY0, int posX1, int posY1)
{
	RECT rectArea = { posX0, posY0, posX1, posY1 };
	if (posX0 == -1)
	{
		rectArea.left = 0;
		rectArea.top = 0;
		rectArea.right = Width;
		rectArea.bottom = Height;
	}
	
	DDBLTFX descFX;
	memset (&descFX, 0, sizeof (descFX));
	descFX.dwSize = sizeof (descFX);
	descFX.dwFillColor = 0;

	CheckDD (Surface->Blt (&rectArea, NULL, NULL, DDBLT_COLORFILL, &descFX), "imgclear");
}
	
/////////////////////////////////
// MImageArray class functions //
/////////////////////////////////

void MImageArray::Create ()
{
	numDib = 0;
	bufDib = NULL;
}
void MImageArray::Destroy ()
{
	int idxDib;

	if (numDib) {
		for (idxDib = 0; idxDib < numDib; idxDib++) bufDib[idxDib].Destroy();
		free (bufDib);
		bufDib = NULL;
		numDib = 0;
	}
}
MImageArray::MImageArray ()
{
	Create ();
}
MImageArray::~MImageArray ()
{
	Destroy ();
}
MIndex MImageArray::GetNextFree()
{
	MIndex idxDib;

	for (idxDib =0; idxDib < numDib; idxDib++) 
		if (bufDib[idxDib].Surface == NULL) return idxDib;
	
	// resize array	
	numDib++;
	bufDib = (MImage *) realloc (bufDib, sizeof (MImage) * numDib);
	bufDib[numDib - 1].Create ();
		
	return numDib - 1;
}
MIndex MImageArray::Add (LPCSTR strPath)
{
	MIndex idxDib;
	
	idxDib = GetNextFree();
	bufDib[idxDib].Load (strPath);
	
	return idxDib;
}
MIndex MImageArray::Add (MImage& mimAdd)
{
	MIndex idxDib;
	
	idxDib = GetNextFree();
	bufDib[idxDib] = mimAdd;
	
	return idxDib;
}
MImage& MImageArray::operator [] (MIndex idxImage)
{
	if (idxImage <0 || idxImage >= numDib) Error("MImageArray.[] :wrong parameter");
	
	return bufDib[idxImage];
}
void MImageArray::Del (MIndex idxImage)
{
	if (idxImage >= numDib || idxImage < 0) Error("MImageArray.Del :wrong parameter");
	if (numDib < 1) 
		Destroy();
	else 
	{
		bufDib[idxImage].Destroy ();
		if (idxImage == numDib - 1) 
		{
			numDib--;
			bufDib = (MImage *) realloc (bufDib, sizeof(MImage) * numDib);
		}	
	}
}

///////////////////////////
// MDraw class functions //
///////////////////////////

void MDraw::Open ()
{
	lpDD = NULL;
	
	CheckDD( DirectDrawCreateEx ( bFix ? ((LPGUID) DDCREATE_EMULATIONONLY) : NULL, (LPVOID *) &lpDD, IID_IDirectDraw7, NULL ), "CreateDD");
	
	CheckDD (lpDD->SetCooperativeLevel( Win->hWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWREBOOT ), "DD Coop");
	
	Screen.Create ();
	ScreenBack.Create ();

	Log ("DirectDraw opened using standard device");
}
void MDraw::Close() 
{
	Screen.Destroy ();
	ScreenBack.Create ();
	
	if (lpDD)
	{
		lpDD->RestoreDisplayMode();
		lpDD->Release();
		lpDD = NULL;
		Screen.Create ();
		Screen.Create ();
	}
}

void MDraw::VideoMode (int newWidth, int newHeight, int newColordepth)
{
	DDSURFACEDESC2 dscSurf;
    DDSCAPS2 capSurf;
	    
	Screen.Destroy ();
	ScreenBack.Create ();

	CheckDD( lpDD->SetDisplayMode( newWidth, newHeight, newColordepth, 0, 0), "DDSetMode");
	
	// Define a flipping surface + backbuffer
	ZeroMemory( &dscSurf, sizeof(DDSURFACEDESC2));
    dscSurf.dwSize = sizeof(DDSURFACEDESC2);
    dscSurf.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
    dscSurf.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP;
    dscSurf.dwBackBufferCount = 1;
	ZeroMemory( &capSurf, sizeof(DDSCAPS2));
	capSurf.dwCaps = DDSCAPS_BACKBUFFER;
    
    // Create the primary surface with 1 back buffer
	CheckDD( lpDD->CreateSurface( &dscSurf, &Screen.Surface, NULL), "Create Surface");
	CheckDD( Screen.Surface->GetAttachedSurface( &capSurf, &ScreenBack.Surface), "Get Back Surface");

	Height = newHeight;
	Width = newWidth;
	Colordepth = newColordepth;
	Screen.Width = Width;
	Screen.Height = Height;
	ScreenBack.Width = Width;
	ScreenBack.Height = Height;

	// clear the screen
	DDBLTFX descFX;
	memset (&descFX, 0, sizeof (descFX));
	descFX.dwSize = sizeof (descFX);
	descFX.dwFillColor = 0;

	CheckDD (Screen.Surface->Blt (NULL, NULL, NULL, DDBLT_COLORFILL, &descFX), "clearscreen");
	CheckDD (ScreenBack.Surface->Blt (NULL, NULL, NULL, DDBLT_COLORFILL, &descFX), "clearscreen");
}

/////////////////////
// Check functions //
/////////////////////

#ifndef NDEBUG

void CheckDD( HRESULT hrFunc, LPCSTR strText) 
{
	if (hrFunc == DD_OK) return;
	
	char strBuf[100];
	sprintf (strBuf, "Direct Draw Error (%lu) : %s", hrFunc, strText);
	Error(strBuf);
}

#endif