// Abiventure app Mgr
// devilducky@gmx.net

#include "tools.h"
#include "filetools.h"
#include "graphtool.h"
#include "math.h"
#include "appmgr.h"

MActor::MActor (LPCSTR strFile, LPCSTR strNameNew)
{
	Create ();
	if (strFile)
		LoadSpecs (strFile);

	if (strNameNew)
	{
		strName = new char [strlen(strNameNew) + 1];
		strcpy (strName, strNameNew);
	}
}
MActor::~MActor ()
{
	Destroy ();
}

void MActor::Create ()
{
	memset (&posCur, 0, sizeof (POINT));
	memset (&rectDest, 0, sizeof (RECT));
	bMove = false;
	strName = NULL;
	bLoaded = false;
	strPath = NULL;
	strDesc = NULL;
	strLook = NULL;
	Text = NULL;
	strDialog = NULL;
	bVisible = true;
	imgFix = NULL;
}

void MActor::Destroy ()
{
	Unload ();

	if (strName)
		delete strName;
	if (strDialog)
		delete strDialog;
	if (strPath)
		delete strPath;
	if (strDesc)
		delete strDesc;
	if (strLook)
		delete strLook;
	if (Text)
		delete Text;
}

void MActor::LoadSpecs (LPCSTR strFile)
{
	TextFile fileActor (strFile);

	// search & attach font
	Text = new MSpeechUnit (App->GetFont (fileActor.FindSetting ("font")), (long) this);
	
	// set path
	strPath = new char [strlen (strFile) + 1];
	strcpy (strPath, strFile);

	bLoaded = false;
}

void MActor::LoadObjSpecs (TextFile &fileRoom, MIndex idxActor)
{
	char strKey[100];
	char strBuffer[256];

	bLoaded = false;

	// name
	sprintf (strKey, "a%d", idxActor);
	if (fileRoom.FindSetting (strKey, strBuffer))
	{
		strName = new char [strlen (strBuffer)+1];
		strcpy (strName, strBuffer);
	}

	// desc
	sprintf (strKey, "a%ddesc", idxActor);
	if (fileRoom.FindSetting (strKey, strBuffer))
	{
		strDesc = new char [strlen (strBuffer)+1];
		strcpy (strDesc, strBuffer);
	}

	// look
	sprintf (strKey, "a%dlook", idxActor);
	if (fileRoom.FindSetting (strKey, strBuffer))
	{
		strLook = new char [strlen (strBuffer)+1];
		strcpy (strLook, strBuffer);
	}

	// file
	sprintf (strKey, "a%dfile", idxActor);
	if (fileRoom.FindSetting (strKey, strBuffer))
	{
		char strFile [256];
		sprintf (strFile, "actor\\%s.act", strBuffer);
		LoadSpecs (strFile);
	}

	// pos
	sprintf (strKey, "a%dpos", idxActor);
	if (fileRoom.FindSetting (strKey, strBuffer))
	{
		SplitVec (strBuffer, &posCur, &dirCur);
	}

	// visible ?
	sprintf (strKey, "a%dinvisible", idxActor);
	if (fileRoom.FindSetting (strKey, NULL, 0))
		bVisible = false;
	
	// 2nd use flag ?
	sprintf (strKey, "a%dusewith", idxActor);
	if (fileRoom.FindSetting (strKey, NULL, 0))
		bUseWith = true;
	
	// composed flag string
	sprintf (strKey, "a%dflag", idxActor);
	if (fileRoom.FindSetting (strKey, strBuffer))
	{
		_strlwr (strBuffer);
		if (!strchr (strBuffer, 'v'))
			bVisible = false;
		if (strchr (strBuffer, 'u'))
			bUseWith = true;
	}

	// dialog
	sprintf (strKey, "a%ddialog", idxActor);
	if (fileRoom.FindSetting (strKey, strBuffer))
	{
		char strFile [256];
		sprintf (strFile, "dialogs\\%s.dlg", strBuffer);
		strDialog = new char [strlen(strFile) + 1];
		strcpy (strDialog, strFile);
	}
}

void MActor::Load ()
{
	if (!bLoaded && strPath)
	{
		TextFile fileActor (strPath);

		memset (posCor, 0, sizeof (POINT) * NUM_ACTORFRAMES);
		memset (&rectDest, 0, sizeof (RECT));
		dirOld = -1;

		// load animation frames	
		for (int iFrame = 0; iFrame < NUM_ACTORFRAMES; iFrame++)
		{
			char strFrameFile[100], strKey[100];
	
			sprintf (strKey, "frame%d", iFrame);
			if (fileActor.FindSetting (strKey))
			{
				sprintf (strFrameFile, "actor\\%s.bmp", fileActor.FindSetting (strKey));
				imgFrame [iFrame].Load (strFrameFile);
			}
			sprintf (strKey, "corx%d", iFrame);
			if (fileActor.FindSetting (strKey))
				posCor [iFrame].x = atoi (fileActor.FindSetting (strKey));
			
			sprintf (strKey, "cory%d", iFrame);
			if (fileActor.FindSetting (strKey))
				posCor [iFrame].y = atoi (fileActor.FindSetting (strKey));
		}

		if (App->bFix && !imgFix)
				imgFix = new MImage (400, 450);

		bLoaded = true;
	}
}

void MActor::Unload ()
{
	if (bLoaded)
	{
		for (int iFrame = 0; iFrame < NUM_ACTORFRAMES; iFrame++)
			imgFrame [iFrame].Destroy ();

		if (imgFix)
		{
			delete imgFix;
			imgFix = NULL;
		}

		bLoaded = false;
	}
}
void MActor::Render (MImage *imgDest)
{
	POINT posCorrected = { posCur.x + posCor [dirCur].x, posCur.y + posCor [dirCur].y };
	RECT rectSource, rectDestOld;

	memcpy (&rectDestOld, &rectDest, sizeof (RECT));

	// scale height
	int iScaledHeight = posZ;
	int iScaledWidth;
	
	if (imgFrame[dirCur].Surface)
		iScaledWidth = posZ * imgFrame[dirCur].Width / imgFrame[dirCur].Height;
	else
		iScaledWidth = posZ * App->CurActor->imgFrame[dirCur].Width / App->CurActor->imgFrame[dirCur].Height;
	
	// target rect
	rectDest.left = posCorrected.x - iScaledWidth / 2;
	rectDest.right = rectDest.left + iScaledWidth;
	rectDest.bottom = posCorrected.y;
	rectDest.top = rectDest.bottom - iScaledHeight;

	// source rect
	rectSource.left = 0;
	rectSource.right = imgFrame[dirCur].Width;
	rectSource.top = 0;
	rectSource.bottom = imgFrame[dirCur].Height;
	
	// clipping
	if (rectDest.left >= 800 || rectDest.right <= 0 || rectDest.bottom <= 0 || rectDest.top >= SCREEN_ROOM_Y )
		return;
	if (rectDest.left < 0)
	{
		rectSource.left -= rectDest.left * imgFrame[dirCur].Width / iScaledWidth;
		rectDest.left = 0;
	}
	if (rectDest.right > 800)
	{
		rectSource.right -= (rectDest.right - 800) * imgFrame[dirCur].Width / iScaledWidth;
		rectDest.right = 800;
	}
	if (rectDest.top < 0)
	{
		rectSource.top -= rectDest.top * imgFrame[dirCur].Height / iScaledHeight;
		rectDest.top = 0;
	}
	if (rectDest.bottom > SCREEN_ROOM_Y)
	{
		rectSource.bottom -= (rectDest.bottom - SCREEN_ROOM_Y) * imgFrame[dirCur].Height / iScaledHeight;
		rectDest.bottom = SCREEN_ROOM_Y;
	}

	if (!imgDest)
		return;

	if (imgFix)
	{
		RECT rectFix = { 0, 0, rectDest.right - rectDest.left, rectDest.bottom - rectDest.top };

		// only update fix buffer if coords change
		if (rectDestOld.top != rectDest.top || rectDestOld.bottom != rectDest.bottom || rectDestOld.left != rectDest.left || rectDestOld.right != rectDest.right || dirOld != dirCur)
		{
			// gdi blit : actorframe to fix buffer
			HDC hdcSrc, hdcDest;
			imgFrame[dirCur].Surface->GetDC (&hdcSrc);
			imgFix->Surface->GetDC (&hdcDest);
			StretchBlt (hdcDest, 0, 0, rectFix.right, rectFix.bottom, hdcSrc, rectSource.left, rectSource.top, rectSource.right - rectSource.left, rectSource.bottom - rectSource.top, SRCCOPY);
			imgFrame[dirCur].Surface->ReleaseDC (hdcSrc);
			imgFix->Surface->ReleaseDC (hdcDest);
		}
		// dx blit : fix buffer to screen
		CheckDD (imgDest->Surface->Blt (&rectDest, imgFix->Surface, &rectFix, DDBLT_KEYSRC | DDBLT_WAIT, NULL), "bltcol");
	}
	else
	{
		CheckDD (imgDest->Surface->Blt (&rectDest, imgFrame[dirCur].Surface, &rectSource, DDBLT_KEYSRC | DDBLT_WAIT, NULL), "actor blit");
	}

	dirOld = dirCur;
}

bool MActor::isMarked (POINT &posMouse)
{
	return (PtInRect (&rectDest, posMouse) != 0);
}

void MActor::WalkTo (POINT posDest, int dirDest, long lID)
{
	lSendID = lID;
	bMove = false;

	if (posDest.x == posCur.x && posDest.y == posCur.y)
	{
		App->Events (CM_ACTORDONE, (WORD) lSendID, (long) this);
		return;
	}

	// check if point is valid
	while (App->CurRoom->ZMap->ZFactor (posDest) == 0)
	{
		posDest.y++;

		// no line found
		if (posDest.y >= SCREEN_ROOM_Y)
			return;
	}
	
	dirGoTo = dirDest;
	
	// start z-trace
	pathCur = App->CurRoom->ZMap->FindPath (posCur, posDest);
		
	if (pathCur && pathCur->timeStamp != 0)
	{
		posCur = pathCur->Pos;
		posZ = pathCur->posZ;
		dirCur = pathCur->Dir;
	
		bMove = true;
	}

}

void MActor::Set (POINT posDest, int dirDest)
{
	posCur = posDest;
	dirCur = dirDest;
	bMove = false;
}

void MActor::UpdateZ ()
{
	posZ = App->CurRoom->ZMap->ZFactor (posCur);
	memset (&rectDest, 0, sizeof (RECT));
}

void MActor::Move (DWORD timeDelta)
{
	if (!bMove)
		return;
	
	pathCur->timeStamp -= timeDelta;
	
	// next one ?
	while (pathCur->timeStamp <= 0)
	{
		long timeOld = pathCur->timeStamp;
		
		// next step
		pathCur++;
		
		// set data
		dirCur = pathCur->Dir;
		posCur = pathCur->Pos;
		posZ = pathCur->posZ;
		
		if (pathCur->timeStamp == -1)
		{
			// finished
			if (dirGoTo != -1)
				dirCur = dirGoTo;

			bMove = false;
			App->Events (CM_ACTORDONE, (WORD) lSendID, (long) this);
			break;
		}

		// carry time over 
		pathCur->timeStamp += timeOld;
	}
}

int MActor::GetDirection (int posDeltaX, int posDeltaY, int iOldDir)
{
	int dirCur;
	
	dirCur = (fabs ((float)posDeltaX) > fabs ((float)posDeltaY * 1.5f)) ? FRAME_LEFT : FRAME_BACK;

	if (dirCur == FRAME_LEFT && posDeltaX > 0)
		dirCur = FRAME_RIGHT;
	
	if (dirCur == FRAME_BACK && posDeltaY > 0)
		dirCur = FRAME_FRONT;

	if ((dirCur & 2) != (iOldDir & 2) && iOldDir >=0)
	{
		if (fabs (atan ((double)posDeltaY / (double)posDeltaX) - 1) < 0.6f)
			dirCur = iOldDir;
	}

	return dirCur;
}

void MActor::RenderText (MImage *imgDest, DWORD timeDelta)
{
	if (Text)
	{
		POINT posText;
		posText.x = posCur.x;
		posText.y = posCur.y;
		
		if (App->CurRoom->ZMap)
			posText.y -= posZ;

		Text->Render (imgDest, timeDelta, &posText);
	}
}

POINT MActor::getWalkPoint (POINT posOther)
{
	POINT posWalk;
	posWalk.y = posCur.y;

	if (posOther.x > posCur.x)
	{
		posWalk.x = posCur.x + TALK_DIST;
		if (posWalk.x >= 800)
			posWalk.x = posCur.x - TALK_DIST;
	}
	else
	{
		posWalk.x = posCur.x - TALK_DIST;
		if (posWalk.x <= 0)
			posWalk.x = posCur.x + TALK_DIST;
	}
	return posWalk;
}

int MActor::getWalkDir (POINT posWalk)
{
	if (posWalk.x > posCur.x)
	{
		return FRAME_LEFT;
	}
	else
	{
		return FRAME_RIGHT;
	}
}

void MActor::setTalkDir (int dirOpposite)
{
	if ((dirOpposite & 3) == FRAME_LEFT)
		dirCur = FRAME_RIGHT;
	else if ((dirOpposite & 3) == FRAME_RIGHT)
		dirCur = FRAME_LEFT;
	else if ((dirOpposite & 3) == FRAME_FRONT)
		dirCur = FRAME_BACK;
	else if ((dirOpposite & 3) == FRAME_BACK)
		dirCur = FRAME_FRONT;
}

void MActor::LoadGame (FILE *fileLoad)
{
	ReadData (fileLoad, posCur);
	ReadData (fileLoad, dirCur);
	ReadData (fileLoad, bVisible);
	ReadData (fileLoad, bUseWith);
	ReadString (fileLoad, strLook);
	
	bMove = false;
	UpdateZ ();
	Render (NULL);
}

void MActor::SaveGame (FILE *fileSave)
{
	WriteData (fileSave, posCur);
	WriteData (fileSave, dirCur);
	WriteData (fileSave, bVisible);
	WriteData (fileSave, bUseWith);
	WriteString (fileSave, strLook);
}