// Mayce wrapper
// devilducky@gmx.net

#include "tools.h"
#include "dxdraw.h"
#include "font.h"

MFont::MFont ()
{
	Create ();
}

MFont::MFont (LPCSTR strFile, LPCSTR strName)
{
	Create ();
	Open (strFile, strName);
}

MFont::~MFont ()
{
	Destroy ();
}

void MFont::Create ()
{
	imgChar = NULL;
	strName = NULL;
	strFile = NULL;
}

void MFont::Destroy ()
{
	if (imgChar)
		delete [] imgChar;

	if (strName)
		delete strName;
	
	if (strFile)
		delete strFile;
	
	Create ();
}

void MFont::Open (LPCSTR strNewFile, LPCSTR strNewName)
{
	if (strNewFile)
	{
		Destroy ();

		// set file, name
		strName = new char [strlen (strNewName) + 1];
		strcpy (strName, strNewName);
		strFile = new char [strlen (strNewFile) + 1];
		strcpy (strFile, strNewFile);
	}
	else
	{
		// reload
		delete [] imgChar;
	}

	imgChar = new MImage [NUM_CHARS];
	
	// mapping keys
	memset (iMapping, -1, 256 * sizeof (int));

	char *strMap = new char [NUM_CHARS + 1];
	strcpy (strMap, MAPPING_STRING);

	
	// build inverse array
	for (MIndex idxChar = 0; idxChar < NUM_CHARS; idxChar++)
	{
		iMapping [(unsigned char) strMap [idxChar]] = idxChar;
	}

	delete [] strMap;

	// color keys
	COLORREF colBookmark = COLOR_BOOKMARK;
	
	// load font image
	MImage imgFont;
	imgFont.Load (strFile);

	// get dc
	HDC hDC;
	imgFont.Surface->GetDC (&hDC);
	
	// check if valid
	Assert0 (GetPixel (hDC, 0, 0) == colBookmark, "fontfile not valid.");

	// process lines
	int posLine = 0, posRow, iHeight, iWidth;
	idxChar = 0;
	bool blnFound = true;
	
	while (true)
	{
		// search height
		blnFound = false;
		for (int posY = posLine + 1; posY < imgFont.Height; posY++)
		{
			if (GetPixel (hDC, 0, posY) == colBookmark)
			{
				blnFound = true;
				iHeight = posY - posLine - 1;
				break;
			}
		}
		
		if (!blnFound)
			break;

		posRow = 0;
		
		// process chars
		while (true)
		{
			// search width
			blnFound = false;
			for (int posX = posRow + 1; posX < imgFont.Width; posX++)
			{
				if (GetPixel (hDC, posX, posLine) == colBookmark)
				{
					blnFound = true;
					iWidth = posX - posRow;
					break;
				}
			}
			if (!blnFound)
				break;

			// alloc space
			imgChar[idxChar].New (iWidth, iHeight);
			
			// free dc , copy char, reget dc
			imgFont.Surface->ReleaseDC (hDC);
			imgChar[idxChar].Blit (0, 0, imgFont, posRow, posLine + 1, iWidth, iHeight);
			imgFont.Surface->GetDC (&hDC);
		
			idxChar ++;
			posRow += iWidth;
		}

		posLine += iHeight + 1;
	}
	
	imgFont.Surface->ReleaseDC (hDC);

	iRowHeight = imgChar[0].Height + ROW_ADD;
}

void MFont::PrintText (int posX, int posY, LPCSTR strText, MImage *imgDest)
{
	for (MIndex idxChar = 0; idxChar < (int) strlen (strText); idxChar++)
	{
		MIndex idxMap = iMapping [(unsigned char) strText [idxChar]];
		
		if (idxMap == -1)
		{
			// space
			posX += LEN_SPACE;
		}
		else
		{
			// set matching key
			imgDest->BlitTransparent (posX, posY, imgChar [idxMap]);
			posX += imgChar [idxMap].Width - 1;
		}
	}
}

void MFont::PrintCenterText (int posX, int posY, LPCSTR strText, MImage *imgDest)
{
	long lenString = 0;
	
	// get string lenght
	for (MIndex idxChar = 0; idxChar < (int) strlen (strText); idxChar++)
	{
		MIndex idxMap = iMapping [(unsigned char) strText [idxChar]];
		
		if (idxMap == -1)
			lenString += LEN_SPACE;
		else
			lenString += imgChar [idxMap].Width - 1;
	}

	// clipping
	int posRealX = posX - lenString / 2;
	int posRealY = posY;
	
	if (posRealX < TEXT_BORDERX)
		posRealX = TEXT_BORDERX;
	if (posRealX > 799 - TEXT_BORDERX - lenString)
		posRealX = 799 - TEXT_BORDERX - lenString;
	if (posRealY < TEXT_BORDERY)
		posRealY = TEXT_BORDERY;

	PrintText (posRealX, posRealY, strText, imgDest);
}