// Abiventure app Mgr
// devilducky@gmx.net

#include "tools.h"
#include "filetools.h"
#include "graphtool.h"
#include "appmgr.h"

MFrame::MFrame ()
{
	Create ();
}
MFrame::~MFrame ()
{
	Destroy ();
}

void MFrame::Create ()
{
	bLoaded = false;
	bVisible = true;
	posX = posY = posZ = 0;
	iType = FRAMETYPE_NONE;
	
	frameNext = NULL;
	strName = NULL;
	strPath = NULL;
}

void MFrame::Destroy ()
{
	if (strName)
		delete [] strName;
	if (strPath)
		delete [] strPath;

	Create ();
}

void MFrame::Load ()
{
	if (strPath[0] && !bLoaded)
	{
		bLoaded = true;
		imgPicture.Load (strPath);
		
		// retrieve z from Zmap
		if (posZ == -1)
		{
			POINT posRoot;
			posRoot.x = posX + imgPicture.Width / 2;
			posRoot.y = posY + imgPicture.Height - 1;

			posZ = App->CurRoom->ZMap->ZFactor (posRoot);
		}
	}
}

void MFrame::Unload ()
{
	if (bLoaded)
		imgPicture.Destroy ();
	
	bLoaded = false;
}

void MFrame::Blit (MImage *imgDest)
{
	imgDest->BlitTransparent (posX, posY, imgPicture);
}

void MFrame::LoadSpecs (TextFile &fileFrames, MIndex idxFrame, MFrame **frameAniStart, MFrame *frameAniNext, RECT rectStretch)
{
	Destroy ();

	char strKey[256], strData[256];
	
	// get name
	sprintf (strKey, "f%d",idxFrame);
	if (fileFrames.FindSetting (strKey, strData))
	{
		strName = new char [strlen (strData) + 1];
		strcpy (strName, strData);
	}

	// get pic data
	sprintf (strKey, "f%dpic", idxFrame);
	strcpy (strData, "frames\\");
	strcat (strData, fileFrames.FindSetting (strKey));

	strPath = new char [strlen (strData) + 1];
	strcpy (strPath, strData);
	
	// is frame visible ?
	sprintf (strKey, "f%dinvisible",idxFrame);
	if (fileFrames.FindSetting (strKey, NULL, 0))
		bVisible = false;

	// composed flag string
	sprintf (strKey, "o%dflag", idxFrame);
	if (fileFrames.FindSetting (strKey, strData))
	{
		_strlwr (strData);
		if (!strchr (strData, 'v'))
			bVisible = false;
	}

	// get type & flags
	sprintf (strKey, "f%dtype", idxFrame);
	fileFrames.FindSetting (strKey, strData);
			
	// check flags
	if (!_stricmp (strData, "back"))
	{
		iType = FRAMETYPE_BACK;
		bVisible = true;
		posZ = -10000;
		if (!strName)
		{
			strName = new char [5];
			strcpy (strName, "back");
		}
	}
	if (!_stricmp (strData, "state"))
	{
		iType = FRAMETYPE_SINGLE;
	}
	if (!_stricmp (strData, "overlay"))
	{
		iType = FRAMETYPE_OVERLAY;
	}
	if (!_strnicmp (strData, "ani", 3))
	{
		iType = FRAMETYPE_ANIMATED;
		bVisible = false;
					
		if (!_stricmp (strData, "aniend"))
			frameNext = frameAniNext;
		else
			frameNext = *frameAniStart;

		if (!_stricmp (strData, "anistart"))
		{
			bVisible = true;
			*frameAniStart = this;
		}
	}

	// get frame pos
	sprintf (strKey, "f%dpos", idxFrame);
	if (fileFrames.FindSetting (strKey, strData))
	{
		posZ = -1;
		SplitVec (strData, &posX, &posY, &posZ);
		
		if (posZ > 0 && rectStretch.left)
		{
			// scale Z pos

			float fScaleFact = ((float)(rectStretch.bottom - rectStretch.top)) / ((float)(rectStretch.right - rectStretch.left));
			float fScale = ((float)rectStretch.top) - ((float)rectStretch.left) * fScaleFact;
			
			posZ = (int) (fScale + fScaleFact * (float) posZ);
		}
	}

	// no name : receive pseudo number
	if (!strName)
	{
		strName = new char [6];
		sprintf (strName, "PF#%d", idxFrame);
	}
}

void MFrame::SaveGame (FILE *fileSave)
{
	WriteData (fileSave, bVisible);
	WriteData (fileSave, posX);
	WriteData (fileSave, posY);
	WriteData (fileSave, posZ);
}

void MFrame::LoadGame (FILE *fileLoad)
{
	ReadData (fileLoad, bVisible);
	ReadData (fileLoad, posX);
	ReadData (fileLoad, posY);
	ReadData (fileLoad, posZ);
}