// Abiventure app Mgr
// devilducky@gmx.net

#include "tools.h"
#include "filetools.h"
#include "graphtool.h"
#include "appmgr.h"

// ITEM CLASS

MItem::MItem ()
{
	Create ();
}
MItem::~MItem ()
{
	Destroy ();
}

void MItem::Create ()
{
	imgPic = NULL;
	strDesc = NULL;
	strLook = NULL;
	strName = NULL;

	bEnabled = false;
	bVisible = false;
	bUseWith = false;
	iType = iState = 0;
}

void MItem::Destroy ()
{
	if (imgPic)
		delete imgPic;
	if (strDesc)
		delete strDesc;
	if (strLook)
		delete strLook;
	if (strName)
		delete strName;
	
	Create ();
}

void MItem::LoadSpecs (TextFile &fileItems, MIndex idxItem)
{
	Destroy ();

	char strKey[256], strData[256];
		
	// get name
	sprintf (strKey, "i%d", idxItem);
	fileItems.FindSetting (strKey, strData);
	strName = new char [strlen (strData) + 1];
	strcpy (strName, strData);

	// get look text
	sprintf (strKey, "i%dlook", idxItem);
	if (fileItems.FindSetting (strKey, strData))
	{
		strLook = new char [strlen (strData) + 1];
		strcpy (strLook, strData);
	}

	// get desc
	sprintf (strKey, "i%ddesc", idxItem);
	if (fileItems.FindSetting (strKey, strData))
	{
		strDesc = new char [strlen (strData) + 1];
		strcpy (strDesc, strData);
	}

	// visible ?
	sprintf (strKey, "i%dvisible", idxItem);
	if (fileItems.FindSetting (strKey, NULL, 0))
		bVisible = true;
	
	// enabled ?
	sprintf (strKey, "i%denabled", idxItem);
	if (fileItems.FindSetting (strKey, NULL, 0))
		bEnabled = true;
	
	// 2nd use flag ?
	sprintf (strKey, "i%dusewith", idxItem);
	if (fileItems.FindSetting (strKey, NULL, 0))
		bUseWith = true;
	
	// composed flag string
	sprintf (strKey, "i%dflag", idxItem);
	if (fileItems.FindSetting (strKey, strData))
	{
		_strlwr (strData);
		if (strchr (strData, 'v'))
			bVisible = true;
		if (strchr (strData, 'e'))
			bEnabled = true;
		if (strchr (strData, 'u'))
			bUseWith = true;
	}

	iType = ITEMTYPE_STD;

	// get type & flags
	sprintf (strKey, "i%dtype", idxItem);
	fileItems.FindSetting (strKey, strData);
			
	// check flags
	if (!_stricmp (strData, "std"))
	{
		iType = ITEMTYPE_STD;
	}

	// load picture
	imgPic = new MImage;

	sprintf (strKey, "items\\item%d.bmp", idxItem);
	imgPic->Load (strKey);
}

// ITEMSTORE CLASS

MItemStore::MItemStore ()
{
	Create ();
}
MItemStore::~MItemStore ()
{
	Destroy ();
}

void MItemStore::Create ()
{
	bRedraw = false;
	numItems = 0;
	Item = NULL;
	idxLine = 0;
	imgBar = NULL;
	rectPos.left = ITEM_X0;
	rectPos.top = ITEM_Y0;
	rectPos.right = ITEM_X0 + ITEMS_LINE * ITEM_WIDTH;
	rectPos.bottom = ITEM_Y0 + 2 * ITEM_HEIGHT;
}

void MItemStore::Destroy ()
{
	if (numItems)
		delete [] Item;
	if (imgBar)
		delete imgBar;

	Create ();
}

void MItemStore::Load (LPCSTR strFile)
{
	Destroy ();

	TextFile fileItems (strFile);

	numItems = atoi (fileItems.FindSetting ("items"));
	Item = new MItem [numItems];

	for (MIndex idxItem = 0; idxItem < numItems; idxItem++)
		Item[idxItem].LoadSpecs (fileItems, idxItem);

	imgBar = new MImage;
	imgBar->Load ("cmdBar\\itembar.bmp");

	bRedraw = true;
}

void MItemStore::Render (MImage *imgDest)
{
	static bool bRenderAgain = true;

	// Render the bar
	imgDest->BlitComplete (ITEMBAR_X, ITEMBAR_Y, *imgBar);	
	
	// Render all items
	MIndex idxPos = 0, idxVis = 0, bVis = false;
	for (MIndex idxItem = 0; idxItem < numItems; idxItem++)
	{
		if (Item[idxItem].bVisible)
		{
			if (idxPos >= idxLine * ITEMS_LINE)
				bVis = true;
			
			if (bVis)
				imgDest->BlitComplete ( ITEM_X0 + ITEM_WIDTH * (idxVis  % ITEMS_LINE), ITEM_Y0 + ITEM_HEIGHT * (idxVis / ITEMS_LINE), *(Item[idxItem].imgPic));

			idxPos++;
			if (bVis)
				idxVis++;

			if (idxVis == 2 * ITEMS_LINE)
				break;
		}
	}

	// Render again (for back surface)
	if (bRenderAgain)
	{
		bRedraw = true;
		bRenderAgain = false;
	}
	else
	{
		bRedraw = false;
		bRenderAgain = true;
	}
}

void MItemStore::Scroll (int iDirection)
{
	idxLine += iDirection;
	
	if (idxLine < 0)
		idxLine = 0;
	if (idxLine * ITEMS_LINE >= numItems)
		idxLine = numItems / ITEMS_LINE;
	
	bRedraw = true;
}

void MItemStore::MouseMove (POINT posMouse)
{
	if (!PtInRect (&rectPos, posMouse))
		return;

	// check all vis. items
	MIndex idxPos = 0, idxVis = 0, bVis = false;
	for (MIndex idxItem = 0; idxItem < numItems; idxItem++)
	{
		if (Item[idxItem].bVisible)
		{
			if (idxPos >= idxLine * ITEMS_LINE)
				bVis = true;
			
			if (bVis)
			{
				int posX = ITEM_X0 + ITEM_WIDTH * (idxVis % ITEMS_LINE);
				int posY = ITEM_Y0 + ITEM_HEIGHT * (idxVis / ITEMS_LINE);

				if (posMouse.x >= posX && posMouse.x < posX + ITEM_WIDTH &&
					posMouse.y >= posY && posMouse.y < posY + ITEM_HEIGHT )
				{
					App->CmdBar->SetMove (Item[idxItem].strDesc);
					return;
				}
			}
			idxPos++;
			if (bVis)
				idxVis++;

			if (idxVis == 2 * ITEMS_LINE)
				break;
		}
	}
}

void MItemStore::Click (POINT posMouse)
{
	// check all vis. items
	MIndex idxPos = 0, idxVis = 0, bVis = false;
	for (MIndex idxItem = 0; idxItem < numItems; idxItem++)
	{
		if (Item[idxItem].bVisible)
		{
			if (idxPos >= idxLine * ITEMS_LINE)
				bVis = true;
			
			if (bVis)
			{
				int posX = ITEM_X0 + ITEM_WIDTH * (idxVis % ITEMS_LINE);
				int posY = ITEM_Y0 + ITEM_HEIGHT * (idxVis / ITEMS_LINE);

				if (posMouse.x >= posX && posMouse.x < posX + ITEM_WIDTH &&
					posMouse.y >= posY && posMouse.y < posY + ITEM_HEIGHT )
				{
					App->CmdBar->SetItem (&Item[idxItem]);
					return;
				}
			}
			idxPos++;
			if (bVis)
				idxVis++;

			if (idxVis == 2 * ITEMS_LINE)
				break;
		}
	}
}

MItem *MItemStore::GetItem (LPCSTR strName)
{
	for (MIndex idxItem = 0; idxItem < numItems; idxItem++)
	{
		if (!_stricmp (Item[idxItem].strName, strName))
			return &Item[idxItem];
	}
	return NULL;
}

void MItemStore::SaveGame (FILE *fileSave)
{
	WriteData (fileSave, idxLine);
	
	for (MIndex idxItem = 0; idxItem < numItems; idxItem++)
	{
		WriteData (fileSave, Item[idxItem].bEnabled);
		WriteData (fileSave, Item[idxItem].bUseWith);
		WriteData (fileSave, Item[idxItem].bVisible);
		WriteString (fileSave, Item[idxItem].strLook);
	}
}

void MItemStore::LoadGame (FILE *fileLoad)
{
	ReadData (fileLoad, idxLine);
	
	for (MIndex idxItem = 0; idxItem < numItems; idxItem++)
	{
		ReadData (fileLoad, Item[idxItem].bEnabled);
		ReadData (fileLoad, Item[idxItem].bUseWith);
		ReadData (fileLoad, Item[idxItem].bVisible);
		ReadDynamicString (fileLoad, Item[idxItem].strLook);
	}
}