// Abiventure app Mgr
// devilducky@gmx.net

#include "tools.h"
#include "filetools.h"
#include "graphtool.h"
#include "appmgr.h"

MObject::MObject ()
{
	Create ();
}
MObject::~MObject ()
{
	Destroy ();
}

void MObject::Create ()
{
	iState = 0;
	iType = 0;
	bEnabled = true;
	bUseWith = true;
	bVisible = true;
	strDesc = NULL;
	strFrame = NULL;
	strName = NULL;
	strDialog = NULL;
	strObj = NULL;
	strLook = NULL;
	Text = NULL;
	memset (&posWalk, 0, sizeof(POINT));
	memset (&posSpeech, 0, sizeof(POINT));
}

void MObject::Destroy ()
{
	if (strDesc)
		delete [] strDesc;
	if (strName)
		delete [] strName;
	if (strLook)
		delete [] strLook;
	if (strFrame)
		delete [] strFrame;
	if (strObj)
		delete [] strObj;
	if (strDialog)
		delete [] strDialog;
	if (Text)
		delete Text;

	Create ();
}

void MObject::LoadSpecs (TextFile &fileObjects, MIndex idxObject)
{
	Destroy ();

	char strKey[256], strData[256];
		
	// get name
	sprintf (strKey, "o%d", idxObject);
	fileObjects.FindSetting (strKey, strData);
	strName = new char [strlen (strData) + 1];
	strcpy (strName, strData);

	// get look text
	sprintf (strKey, "o%dlook", idxObject);
	if (fileObjects.FindSetting (strKey, strData))
	{
		strLook = new char [strlen (strData) + 1];
		strcpy (strLook, strData);
	}

	// get desc
	sprintf (strKey, "o%ddesc", idxObject);
	if (fileObjects.FindSetting (strKey, strData))
	{
		strDesc = new char [strlen (strData) + 1];
		strcpy (strDesc, strData);
	}

	// get dialog
	sprintf (strKey, "o%ddialog", idxObject);
	if (fileObjects.FindSetting (strKey, strData))
	{
		char strFile[256];
		sprintf (strFile, "dialogs\\%s.dlg", strData);
		strDialog = new char [strlen (strFile) + 1];
		strcpy (strDialog, strFile);
	}

	// get obj attachment
	sprintf (strKey, "o%dattach", idxObject);
	if (fileObjects.FindSetting (strKey, strData))
	{
		strObj = new char [strlen (strData) + 1];
		strcpy (strObj, strData);
	}

	// get frame attachment
	sprintf (strKey, "o%dframe", idxObject);
	if (fileObjects.FindSetting (strKey, strData))
	{
		strFrame = new char [strlen (strData) + 1];
		strcpy (strFrame, strData);
	}

	// visible ?
	sprintf (strKey, "o%dinvisible", idxObject);
	if (fileObjects.FindSetting (strKey, NULL, 0))
		bVisible = false;
	
	// enabled ?
	sprintf (strKey, "o%ddisabled", idxObject);
	if (fileObjects.FindSetting (strKey, NULL, 0))
		bEnabled = false;
	
	// 2nd use flag ?
	sprintf (strKey, "o%dusesingle", idxObject);
	if (fileObjects.FindSetting (strKey, NULL, 0))
		bUseWith = false;
	
	// composed flag string
	sprintf (strKey, "o%dflag", idxObject);
	if (fileObjects.FindSetting (strKey, strData))
	{
		bVisible = bEnabled = bUseWith = false;

		_strlwr (strData);
		if (strchr (strData, 'v'))
			bVisible = true;
		if (strchr (strData, 'e'))
			bEnabled = true;
		if (strchr (strData, 'u'))
			bUseWith = true;
	}

	// get type & flags
	sprintf (strKey, "o%dtype", idxObject);
	fileObjects.FindSetting (strKey, strData);
			
	// check flags
	if (!_stricmp (strData, "std"))
	{
		iType = OBJECTTYPE_STD;
	}
	if (!_stricmp (strData, "door"))
	{
		iType = OBJECTTYPE_DOOR;
	}
	if (!_stricmp (strData, "item"))
	{
		iType = OBJECTTYPE_ITEM;
	}
	if (!_stricmp (strData, "talk"))
	{
		iType = OBJECTTYPE_TALK;
		bVisible = false;
	}
	if (!_stricmp (strData, "person"))
	{
		iType = OBJECTTYPE_PERSON;
	}
	if (!_stricmp (strData, "group"))
	{
		iType = OBJECTTYPE_GROUP;
	}

	// get frame pos
	sprintf (strKey, "o%dpos", idxObject);
	if (fileObjects.FindSetting (strKey, strData))
	{
		SplitRect (strData, &rectPos);
	}

	// get speech pos
	sprintf (strKey, "o%dspeech", idxObject);
	if (fileObjects.FindSetting (strKey, strData))
	{
		SplitVec (strData, &posSpeech);
	}

	// get font & init speech unit
	sprintf (strKey, "o%dfont", idxObject);
	if (fileObjects.FindSetting (strKey, strData))
	{
		Text = new MSpeechUnit (App->GetFont (strData), (long)this);
	}

	// get walk pos
	sprintf (strKey, "o%dwalk", idxObject);
	if (fileObjects.FindSetting (strKey, strData))
	{
		dirWalk = -1;
		SplitVec (strData, &posWalk, &dirWalk);
	}
	else
	{
		posWalk.x = (rectPos.left + rectPos.right) / 2;
		posWalk.y = rectPos.bottom;
		dirWalk = -1;
	}
}

void MObject::SaveGame (FILE *fileSave)
{
	WriteData (fileSave, bVisible);
	WriteData (fileSave, bUseWith);
	WriteData (fileSave, bEnabled);
	WriteString (fileSave, strLook);
}

void MObject::LoadGame (FILE *fileLoad)
{
	ReadData (fileLoad, bVisible);
	ReadData (fileLoad, bUseWith);
	ReadData (fileLoad, bEnabled);
	ReadDynamicString (fileLoad, strLook);
}
