// Abiventure app Mgr
// devilducky@gmx.net

#include "tools.h"
#include "filetools.h"
#include "graphtool.h"
#include "appmgr.h"

MRoom::MRoom ()
{
	Create ();
}
MRoom::~MRoom ()
{
	Destroy ();
}

void MRoom::Create ()
{
	numFrames = 0;
	Frame = NULL;
	numObjects = 0;
	Object = NULL;
	numActors = 0;
	Actor = NULL;
	CurActor = NULL;
	rectPos.top = 0;
	rectPos.left = 0;
	rectPos.bottom = SCREEN_ROOM_Y;
	rectPos.right = 800;
	ZMap = NULL;
	strName = NULL;
	numUnits = 0;
	Unit = NULL;
	bBlack = false;
}

void MRoom::Destroy ()
{
	if (numFrames)
		delete [] Frame;
	
	if (numObjects)
		delete [] Object;
		
	if (numActors)
	{
		for (MIndex idxActor = 0; idxActor < numActors; idxActor++)
			delete Actor[idxActor];

		free (Actor);
	}

	if (ZMap)
		delete ZMap;
	if (strName)
		delete strName;
	if (Unit)
		delete [] Unit;

	Create ();
}

void MRoom::Load (LPCSTR strFile)
{
	Destroy ();

	TextFile fileRoom (strFile);
	
	// get name
	char strBuffer[100];
	fileRoom.FindSetting ("name", strBuffer);
	strName = new char[strlen (strBuffer) + 1];
	strcpy (strName, strBuffer);

	// reserve memory
	numFrames = atoi (fileRoom.FindSetting ("frames"));
	Frame = new MFrame [numFrames];
	
	numActors = atoi (fileRoom.FindSetting ("actors"));
	Actor = (MActor **) malloc (numActors * sizeof (MActor *));
	
	numObjects = atoi (fileRoom.FindSetting ("objects"));
	Object = new MObject [numObjects];

	RECT rectStretch;
	memset (&rectStretch, 0, sizeof(rectStretch));
	dirStart = 0;
	ZMap = NULL;
		
	// flatmap ??
	if (fileRoom.FindSetting ("flatmap"))
	{	
		// get text out
		SplitVec (fileRoom.FindSetting ("flatmap"), &posStart);
	}
	else 
	{
		// get scale factor
		SplitRect (fileRoom.FindSetting ("scale"), &rectStretch);
		if (rectStretch.right > rectStretch.left)
		{
			SwapInt (rectStretch.left, rectStretch.right);
			SwapInt (rectStretch.bottom, rectStretch.top);
		}
		
		// set curactor startpos
		SplitVec (fileRoom.FindSetting ("start"), &posStart, &dirStart );
	
		// init zmap
		char strSetting[256];

		fileRoom.FindSetting ("zmap", strSetting);

		if (strSetting[0])
		{
			char strPath[256];
			sprintf (strPath, "zmap\\%s", strSetting);
		
			ZMap = new MZMap ();
			ZMap->Set (strPath, rectStretch);
		}	
	}

	// load frames
	MFrame *frameAniStart;
	for (MIndex idxFrame = 0; idxFrame < numFrames; idxFrame++)
	{
		Frame[idxFrame].LoadSpecs (fileRoom, idxFrame, &frameAniStart, &Frame[idxFrame + 1], rectStretch);
	}

	// load objects
	for (MIndex idxObject = 0; idxObject < numObjects; idxObject++)
	{
		Object[idxObject].LoadSpecs (fileRoom, idxObject);
	}

	// load actors
	for (MIndex idxActor = 0; idxActor < numActors; idxActor++)
	{
		Actor[idxActor] = new MActor;
		Actor[idxActor]->LoadObjSpecs (fileRoom, idxActor);
	}
}

void MRoom::Enter (bool bCauseEvent)
{
	if (ZMap)
		ZMap->Load ();

	for (MIndex idxFrame = 0; idxFrame < numFrames; idxFrame++)
	{
		// load bitmaps in memory
		Frame[idxFrame].Load ();
	}
	for (MIndex idxActor = 0; idxActor < numActors; idxActor++)
	{
		// load actor frames in memory
		Actor[idxActor]->Load ();
		Actor[idxActor]->UpdateZ ();
		Actor[idxActor]->Render (NULL);
		if (Actor[idxActor]->Text)
			Actor[idxActor]->Text->Cancel ();
	}

	if (ZMap)
	{
		// set actor to start pos & direction
		if (CurActor)
		{
			CurActor->bVisible = true;
			CurActor->Set (posStart, dirStart);
			CurActor->UpdateZ ();
			CurActor->Text->Cancel ();
		}
	}
	else
	{
		CurActor->bVisible = false;
		CurActor->Set (posStart, 0);
	}

	// startup event
	if (bCauseEvent)
	{
		LPCSTR strArg[2] = { NULL, NULL };
		App->RoomMgr->DoCmd (this, CMD_ENTER, strArg);
	}
}

void MRoom::Leave (bool bCauseEvent)
{
	// save actor pos
	memcpy (&posStart, &CurActor->posCur, sizeof (POINT));
	dirStart = CurActor->dirCur & FRAME_MASK;
	
	if (bCauseEvent)
	{
		switch (dirStart)
		{
		case FRAME_LEFT:
			dirStart = FRAME_RIGHT; break;
		case FRAME_RIGHT:
			dirStart = FRAME_LEFT; break;
		case FRAME_FRONT:
			dirStart = FRAME_BACK; break;
		case FRAME_BACK:
			dirStart = FRAME_FRONT; break;
		}
		
		// termination event

		LPCSTR strArg[2] = { NULL, NULL };
		App->RoomMgr->DoCmd (this, CMD_LEAVE, strArg);
	}

	for (MIndex idxFrame = 0; idxFrame < numFrames; idxFrame++)
	{
		// free bitmaps
		Frame[idxFrame].Unload ();
	}
	for (MIndex idxActor = 0; idxActor < numActors; idxActor++)
	{
		// free actor frames
		Actor[idxActor]->Unload ();
	}

	if (ZMap)
		ZMap->Unload ();
}

void MRoom::Render (MImage *imgDest, DWORD timeDelta)
{
	int posZ;
	MIndex idxNext;
	
	// move actors
	for (MIndex idxActor = 0; idxActor < numActors; idxActor++)
		Actor[idxActor]->Move (timeDelta);
	
	if (!bBlack)
	{
		// empty done array
		bool *bFrameDone = new bool [numFrames];
		memset (bFrameDone, 0, sizeof(bool) * numFrames);
		
		bool *bActorDone = new bool [numActors + 1];
		memset (bActorDone, 0, sizeof(bool) * (numActors + 1));
		
		do
		{
			// get farest frame
			posZ = 32000;
			idxNext = -1;

			for (MIndex idxFrame = 0; idxFrame < numFrames; idxFrame++)
			{
				if (!bFrameDone [idxFrame] && Frame[idxFrame].bVisible && Frame[idxFrame].posZ <= posZ)
				{
					posZ = Frame [idxFrame].posZ;
					idxNext = idxFrame;
				}
			}
			// get farest actor
			for (MIndex idxActor = 0; idxActor < numActors; idxActor++)
			{
				if (!bActorDone [idxActor] && Actor[idxActor]->bVisible && Actor[idxActor]->posZ <= posZ)
				{
					posZ = Actor [idxActor]->posZ;
					idxNext = -2 - idxActor;
				}
			}

			if (CurActor && CurActor->bVisible && !bActorDone[numActors] && CurActor->posZ <= posZ)
			{
				CurActor->Render (imgDest);
				bActorDone [numActors] = true;
			}
			else if (idxNext == -1)
			{
				break;
			}
			else
			{
				if (idxNext >= 0)
				{
					bFrameDone [idxNext] = true;
					if (Frame [idxNext].bVisible)
						Frame [idxNext].Blit (imgDest);
				}
				else 
				{
					idxNext = -2 - idxNext;
					bActorDone [idxNext] = true;
					Actor [idxNext]->Render (imgDest);
				}
			}
		} while (true);
	}
	else
	{
		// Blackness 
		imgDest->Clear (0, 0, 800, SCREEN_ROOM_Y);
	}

	// render actor text
	for (idxActor = 0; idxActor < numActors; idxActor++)
		Actor[idxActor]->RenderText (imgDest, timeDelta);
	
	// render object-actor text
	for (MIndex idxObject = 0; idxObject < numObjects; idxObject++)
		if (Object[idxObject].Text)
			Object[idxObject].Text->Render (imgDest, timeDelta, &Object[idxObject].posSpeech);
	
	// render additional speech units
	for (MIndex idxUnit = 0; idxUnit < numUnits; idxUnit++)
		Unit[idxUnit]->Render (imgDest, timeDelta);
}

void MRoom::Click (POINT &posMouse)
{
	for (MIndex idxActor = 0; idxActor < numActors; idxActor++)
	{
		if (Actor[idxActor]->bVisible && Actor[idxActor]->isMarked (posMouse))
		{
			App->CmdBar->SetActor (Actor[idxActor]);
			return;
		}
	}
	for (MIndex idxObject = 0; idxObject < numObjects; idxObject++)
	{
		if (Object[idxObject].bVisible && PtInRect (&Object[idxObject].rectPos, posMouse))
		{
			App->CmdBar->SetObject (&Object[idxObject]);
			return;
		}
	}
	

	// goto 
	if (App->CmdBar->bRunning)
		App->CmdBar->Cancel ();

	if (ZMap && CurActor)
		CurActor->WalkTo (posMouse);
}

void MRoom::MouseMove (POINT &posMouse)
{
	if (posMouse.y > SCREEN_ROOM_Y)
		return;

	for (MIndex idxActor = 0; idxActor < numActors; idxActor++)
	{
		if (Actor[idxActor]->bVisible && Actor[idxActor]->isMarked (posMouse))
		{
			App->CmdBar->SetMove (Actor[idxActor]->strDesc);
			return;
		}
	}
	for (MIndex idxObject = 0; idxObject < numObjects; idxObject++)
	{
		if (Object[idxObject].bVisible && PtInRect (&Object[idxObject].rectPos, posMouse))
		{
			App->CmdBar->SetMove (Object[idxObject].strDesc);
			return;
		}
	}
}

MFrame *MRoom::getFrame (LPCSTR strName)
{
	for (MIndex idxFrame = 0; idxFrame < numFrames; idxFrame++)
	{
		if (!_stricmp (Frame[idxFrame].strName, strName))
			return &Frame[idxFrame];
	}
	return NULL;
}

MActor *MRoom::getActor (LPCSTR strName)
{
	for (MIndex idxActor = 0; idxActor < numActors; idxActor++)
	{
		if (!_stricmp (Actor[idxActor]->strName, strName))
			return Actor[idxActor];
	}
	return NULL;
}

MObject *MRoom::getObject (LPCSTR strName)
{
	for (MIndex idxObject = 0; idxObject < numObjects; idxObject++)
	{
		if (!_stricmp (Object[idxObject].strName, strName))
			return &Object[idxObject];
	}
	return NULL;
}

void MRoom::AttachUnit (MSpeechUnit *speAttach)
{
	numUnits++;
	Unit = (MSpeechUnit **) realloc (Unit, numUnits * sizeof (MSpeechUnit *));
	
	Unit [numUnits - 1] = speAttach;
}

void MRoom::DetachUnit (MSpeechUnit *speDetach)
{
	for (MIndex idxUnit = 0; idxUnit < numUnits; idxUnit++)
	{
		if (speDetach == Unit [idxUnit])
		{
			// remove the Unit & push the following 1 forward
			memmove (&Unit [idxUnit], &Unit [idxUnit + 1], (numUnits - idxUnit - 1) * sizeof (MSpeechUnit *));
			numUnits--;
			Unit = (MSpeechUnit **) realloc (Unit, numUnits * sizeof (MSpeechUnit *));
			break;
		}
	}
}

void MRoom::AttachActor (MActor *actAttach)
{
	numActors++;
	Actor = (MActor **) realloc (Actor, numActors * sizeof (MActor *));
	
	Actor [numActors - 1] = actAttach;
}

void MRoom::DetachActor (MActor *actDetach)
{
	for (MIndex idxActor = 0; idxActor < numActors; idxActor++)
	{
		if (actDetach == Actor [idxActor])
		{
			// remove the actor & push the following 1 forward
			memmove (&Actor [idxActor], &Actor [idxActor + 1], (numActors - idxActor - 1) * sizeof (MActor *));
			numActors--;
			Actor = (MActor **) realloc (Actor, numActors * sizeof (MActor *));
			break;
		}
	}
}

void MRoom::LoadGame (FILE *fileLoad, int iVersion)
{
	ReadData (fileLoad, posStart);
	ReadData (fileLoad, dirStart);
	ReadData (fileLoad, bBlack);

	int numNewObjects = numObjects;
	int numNewFrames = numFrames;
	int numNewActors = numActors;
	
	// Backward compatibility
	if (iVersion < 151 && !_stricmp (strName, "schulhof"))
	{
		numNewObjects--;
		numNewActors--;
	}

	for (MIndex idxObj = 0; idxObj < numNewObjects; idxObj++)
		Object[idxObj].LoadGame (fileLoad);
	
	for (MIndex idxFrame = 0; idxFrame < numNewFrames; idxFrame++)
		Frame[idxFrame].LoadGame (fileLoad);
	
	for (MIndex idxActor = 0; idxActor < numNewActors; idxActor++)
		Actor[idxActor]->LoadGame (fileLoad);
}

void MRoom::SaveGame (FILE *fileSave)
{
	WriteData (fileSave, posStart);
	WriteData (fileSave, dirStart);
	WriteData (fileSave, bBlack);

	for (MIndex idxObj = 0; idxObj < numObjects; idxObj++)
		Object[idxObj].SaveGame (fileSave);
	
	for (MIndex idxFrame = 0; idxFrame < numFrames; idxFrame++)
		Frame[idxFrame].SaveGame (fileSave);
	
	for (MIndex idxActor = 0; idxActor < numActors; idxActor++)
		Actor[idxActor]->SaveGame (fileSave);
}

void MRoom::Skip ()
{
	// skip actor text
	for (MIndex idxActor = 0; idxActor < numActors; idxActor++)
		Actor[idxActor]->Text->Skip ();
	
	// skip object-actor text
	for (MIndex idxObject = 0; idxObject < numObjects; idxObject++)
		if (Object[idxObject].Text)
			Object[idxObject].Text->Skip ();
	
	// skip additional speech units
	for (MIndex idxUnit = 0; idxUnit < numUnits; idxUnit++)
		Unit[idxUnit]->Skip ();
}