// Abiventure app Mgr
// devilducky@gmx.net

#include "tools.h"
#include "mmsystem.h"
#include "filetools.h"
#include "graphtool.h"
#include "appmgr.h"

#define TIMING_AHEAD	20
#define CURVE_LI_CHECK	10
#define MOVE_WAIT_XY	2500.0f
#define MOVE_WAIT_Z		750.0f
#define ACTOR_STEP		240.0f

MZMap::MZMap ()
{
	Create ();
}

MZMap::~MZMap ()
{
	Destroy ();
}

void MZMap::Create ()
{
	ZData = NULL;
	strPath = NULL;
	pathStep = NULL;
	bLoaded = false;
}

void MZMap::Destroy ()
{
	if (ZData)
		delete [] ZData;
	if (strPath)
		delete [] strPath;
	if (pathStep)
		delete [] pathStep;
	
	Create ();
}

void MZMap::Set (LPCSTR strNewFile, RECT rectNewStretch)
{
	Destroy ();

	strPath = new char [strlen (strNewFile) + 1];
	strcpy (strPath, strNewFile);

	rectStretch = rectNewStretch;
}

void MZMap::Unload ()
{
	if (ZData)
		delete [] ZData;
	if (pathStep)
		delete [] pathStep;

	ZData = NULL;
	pathStep = NULL;

	bLoaded = false;
}

void MZMap::Load ()
{
	HBITMAP hBitmap;
	BITMAP bmpZMap;

	// destroy old data
	Unload ();

	// Load the zmap
	Check0 (hBitmap = (HBITMAP) LoadImage (NULL, strPath, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE | LR_CREATEDIBSECTION), "load zmap");
	
	// Get bitmap data
	GetObject(hBitmap, sizeof(BITMAP), &bmpZMap);

	Assert0 (bmpZMap.bmBitsPixel == 8, "wrong zmap type (should be 256 grey)");
    
	// calculate strech factors
	
	float fScaleFact = ((float)(rectStretch.bottom - rectStretch.top)) / ((float)(rectStretch.right - rectStretch.left));
	float fScale = ((float)rectStretch.top) - ((float)rectStretch.left) * fScaleFact;
	
	int numLines = bmpZMap.bmHeight < SCREEN_ROOM_Y ? bmpZMap.bmHeight : SCREEN_ROOM_Y;
	
	// new z-array
	ZData = new int [SCREEN_ROOM_Y * 800];
	memset (ZData, 0, sizeof (int) * SCREEN_ROOM_Y * 800);

	// copy pixel data
	for (int posY = 0; posY < numLines; posY++)
	{
		for (int posX = 0; posX < 800; posX++)
		{
			int iData = ((LPBYTE) bmpZMap.bmBits) [posX + (numLines - posY) * 800];

			if (iData)
				ZData [posX + posY * 800] = ((int) (fScale + (float) iData * fScaleFact));
		}
	}

	DeleteObject (hBitmap);

	bLoaded = true;
}

MPathData * MZMap::FindPath (POINT posFrom, POINT posTo)
{
	// destroy old path
	if (pathStep)
		delete [] pathStep;

	pathStep = NULL;

	int iSteps;
	
	// try different strategies :
	// first : linear interpolation

	if ((iSteps = Interpolate (posFrom, posTo, NULL)) > 0 )
	{
		// found path
		pathStep = new MPathData [iSteps];
		Interpolate (posFrom, posTo, pathStep);
		Timing (iSteps);
		return pathStep;		
	}
	
	// Then, use trace algorithm
	if ((iSteps = TraceCurves (posFrom, posTo)) > 0 )
	{
		Timing (iSteps);
		return pathStep;
	}

	return NULL;
}

int MZMap::Interpolate (POINT posFrom, POINT posTo, MPathData *pathWrite)
{
	MInterpolation ipolPath;
	int iSteps = 0;
	int posZ;
	
	ipolPath.Start (posFrom.x, posFrom.y, posTo.x, posTo.y);
	
	// step loop
	do
	{
		iSteps++;
		
		posZ = ZFactor (ipolPath.posCur);

		if (pathWrite)
		{
			// Write position
			pathWrite[iSteps - 1].Pos = ipolPath.posCur;
			pathWrite[iSteps - 1].posZ = posZ;
		}
		if (posZ == 0)
		{
			// path invalid
			return -iSteps;
		}

		ipolPath.Frame ();
	}
	while (!(ipolPath.posCur.x == posTo.x && ipolPath.posCur.y == posTo.y ));

	// terminal step
	iSteps++;
	if (pathWrite)
	{
		pathWrite[iSteps - 1].Pos = posTo;
		pathWrite[iSteps - 1].posZ = posZ;
	}
	return iSteps;
}

void MZMap::Timing (int iSteps)
{
	float fFrame = 0;
	int iOldDir = -1;
	int iState = FRAME_STEP1;
	int iStepState = 1;
	
	for (MIndex idxStep = 0; idxStep < iSteps - 1; idxStep+= TIMING_AHEAD)
	{
		// interpolate 10 steps forward
		int iNextStep = (idxStep + TIMING_AHEAD >= iSteps) ? (iSteps - 1) : (idxStep + TIMING_AHEAD);
		
		// get way difference
		int posDeltaX, posDeltaY, posDeltaZ;
		posDeltaX = pathStep[iNextStep].Pos.x - pathStep[idxStep].Pos.x;
		posDeltaY = pathStep[iNextStep].Pos.y - pathStep[idxStep].Pos.y;
		posDeltaZ = ZFactor (pathStep[iNextStep].Pos) - ZFactor (pathStep[idxStep].Pos);
		
		// direction & frame
		int iDir = App->CurActor->GetDirection (posDeltaX, posDeltaY, iOldDir);
		iOldDir = iDir;
		
		// calculate wait time
		if (posDeltaX < 0) posDeltaX = -posDeltaX;
		if (posDeltaY < 0) posDeltaY = -posDeltaY;
		if (posDeltaZ < 0) posDeltaZ = -posDeltaZ;

		float fWaitTime = ((float) (posDeltaX + posDeltaY) * MOVE_WAIT_XY + (float) posDeltaZ * MOVE_WAIT_Z) / (float)(iNextStep - idxStep);
		
		// calc block
		for (MIndex idxBlock = idxStep; idxBlock < iNextStep; idxBlock++)
		{
			// calc zpos
			pathStep[idxBlock].posZ = ZFactor (pathStep[idxBlock].Pos);
		
			// direction & frame
			pathStep[idxBlock].Dir = iDir | iState;

			// time
			pathStep[idxBlock].timeStamp = (long) (fWaitTime / (float) pathStep[idxBlock].posZ);
		
			// get framestep state
			fFrame += pathStep[idxBlock].timeStamp;
			if (fFrame > ACTOR_STEP)
			{
				fFrame -= ACTOR_STEP;
				iStepState++;
				iStepState &= 3;

				switch (iStepState)
				{
				case 0:
				case 2:
					iState = 0;
					break;
				case 1:
					iState = FRAME_STEP1;
					break;
				case 3:
					iState = FRAME_STEP2;
					break;
				}
			}
		}
	}
	
	pathStep [iSteps - 1].timeStamp = -1;
	pathStep [iSteps - 1].posZ = ZFactor (pathStep[iSteps - 1].Pos);
	pathStep [iSteps - 1].Dir = pathStep [iSteps - 2].Dir & 3;
}

int MZMap::TraceCurves (POINT posFrom, POINT posTo)
{
	// get some space
	int iReservedSteps;
	int iSteps;
	int iStepsCurve[2];
	MPathData *pathCurve[2];
	MPathData pathLast[5];
		
	for (MIndex dirPath = 0; dirPath < 2; dirPath++)
	{
		// get some mem
		iReservedSteps = 1000;
		MPathData *pathTry = (MPathData *) malloc (sizeof (MPathData) * iReservedSteps);
		
		// do linear interpolation until barrier hit
		iSteps = -Interpolate (posFrom, posTo, pathTry);

		if (iSteps < 1)
		{
			iSteps = 1;
			pathTry[0].Pos = posFrom;
		}

		// set li start bookmark
		pathTry[0].posZ = -1;

		// first lastpath entries
		memset (pathLast, 0, sizeof (pathLast));
		if (iSteps >= 5)
			memcpy (pathLast, &pathTry[iSteps - 5], 5 * sizeof (MPathData));
		else
			memcpy (pathLast, pathTry, iSteps * sizeof (MPathData));

		// direction check flag ( at start only )
		bool bCheckDir = true;
		int iLinearTest = 0;

		// trace the outline
		while (1)
		{
			iLinearTest++;
			if (iLinearTest == CURVE_LI_CHECK)
				iLinearTest = 0;

			// first, try to establish linear connection
			if (!iLinearTest && Interpolate (pathTry[iSteps - 1].Pos, posTo, NULL) > 0)
			{
				// final step !
				// get some mem and finalize
				iReservedSteps = iSteps + 1000;
				pathTry = (MPathData *) realloc (pathTry, sizeof (MPathData) * iReservedSteps);

				iSteps += Interpolate (pathTry[iSteps - 1].Pos, posTo, &pathTry [iSteps - 1]) - 1;
				break;
			}
			
			bool bChanged = false;
			int dirFirst = -1;
			POINT posRequire = { 0,0 };
			
			// if Checkflag is enabled, find possible direction points
			if (bCheckDir)
			{
				for (MIndex idxCurField = 0; idxCurField < 9; idxCurField++)
				{
					if (idxCurField == 4) continue;

					POINT posTest;
					posTest.x = pathTry[iSteps - 1].Pos.x + (idxCurField % 3) - 1;
					posTest.y = pathTry[iSteps - 1].Pos.y + (idxCurField / 3) - 1;
					if (posTest.x < 0 || posTest.y < 0 || posTest.x >= 800 || 
						posTest.y >= SCREEN_ROOM_Y || ZFactor (posTest) != 0)
						continue;
					
					// test checkflag !
					bool bCount = false;

					// check to surrounding fields
					switch (idxCurField)
					{
					case 1:
					case 7: 
						if (ZData [posTest.x + 1 + posTest.y * 800] != 0 || ZData [posTest.x - 1 + posTest.y * 800] != 0)
							bCount = true;
						break;
					case 3:
					case 5:
						if (ZData [posTest.x + (posTest.y + 1) * 800] != 0 || ZData [posTest.x + (posTest.y - 1) * 800] != 0)
							bCount = true;
						break;
					case 0:
						if (ZData [posTest.x + 1 + posTest.y * 800] != 0 || ZData [posTest.x + (posTest.y + 1) * 800] != 0)
							bCount = true;
						break;
					case 2:
						if (ZData [posTest.x - 1 + posTest.y * 800] != 0 || ZData [posTest.x + (posTest.y + 1) * 800] != 0)
							bCount = true;
						break;
					case 6:
						if (ZData [posTest.x + 1 + posTest.y * 800] != 0 || ZData [posTest.x + (posTest.y - 1) * 800] != 0)
							bCount = true;
						break;
					case 8:
						if (ZData [posTest.x - 1 + posTest.y * 800] != 0 || ZData [posTest.x + (posTest.y - 1) * 800] != 0)
							bCount = true;
						break;
					}

					if (bCount)
					{
						// only 1 field found -> cont. search
						// 2 fields -> get dir, disable check flag
						if (dirFirst == -1)
						{
							dirFirst = idxCurField;
						}
						else
						{
							bCheckDir = false;
							if (dirPath)
								dirFirst = idxCurField;
							
							// set matching direction
							posRequire.x = pathTry[iSteps - 1].Pos.x + (dirFirst % 3) - 1;
							posRequire.y = pathTry[iSteps - 1].Pos.y + (dirFirst / 3) - 1;
							break;
						}
					}	
				}
			}	
			
			// find the next trace-point ( skip middle, points on the border )
			for (MIndex idxCurField = 0; idxCurField < 9; idxCurField++)
			{
				if (idxCurField == 4) continue;

				POINT posTest;
				posTest.x = pathTry[iSteps - 1].Pos.x + (idxCurField % 3) - 1;
				posTest.y = pathTry[iSteps - 1].Pos.y + (idxCurField / 3) - 1;
				if (posTest.x < 0 || posTest.y < 0 || posTest.x >= 800 || 
					posTest.y >= SCREEN_ROOM_Y || ZFactor (posTest) == 0) continue;

				bool bValid = false;
				
				// test : is that point near the border ?
				for (MIndex idxNextField = 0; idxNextField < 9; idxNextField++)
				{
					if (idxNextField == 4) continue;
					
					POINT posNext;
					posNext.x = posTest.x + (idxNextField % 3) - 1;
					posNext.y = posTest.y + (idxNextField / 3) - 1;
					if (posNext.x < 0 || posNext.y < 0 || posNext.x >= 800 || 
						posNext.y >= SCREEN_ROOM_Y || ZFactor (posNext) != 0) continue;

					if (posRequire.x != 0 && posRequire.y != 0 && (posNext.x != posRequire.x || posNext.y != posRequire.y))
						continue;
					
					bValid = true;
					break;
				}
				
				// test : was this point already listed ? ( last 5 )
				if (!posRequire.x && !posRequire.y)
				{
					for (MIndex idxPrev = 0; idxPrev < 5; idxPrev++)
					{
						if (pathLast[idxPrev].Pos.x == posTest.x && pathLast[idxPrev].Pos.y == posTest.y)
						{
							// already listed---
							bValid = false;
							break;
						}
					}
				}
				if (bValid)
				{
					// strike, gotcha !
					// add waypoint
					iSteps++;
					if (iSteps > iReservedSteps)
					{
						iReservedSteps += 1000;
						pathTry = (MPathData *) realloc (pathTry, sizeof (MPathData) * iReservedSteps);
					}
					pathTry[iSteps - 1].Pos = posTest;
					pathTry[iSteps - 1].posZ = -1;
					bChanged = true;

					// add last point
					memmove (&pathLast[1], &pathLast[0], 4 * sizeof (MPathData));
					memcpy (&pathLast[0], &pathTry[iSteps - 1], sizeof(MPathData));
					
					break;
				}
			}
			
			if (!bChanged)
			{
				// failed
				iSteps = 30000;
				break;
			}
			
			if (iLinearTest)
				continue;

			// test if new point could have been reached by LI ( shortcut ! )
			for (MIndex idxPrev = 0; idxPrev < iSteps - 5; idxPrev ++)
			{
				if (pathTry [idxPrev].posZ != -1)
					continue;

				int iNewSteps;
				if ((iNewSteps = Interpolate (pathTry[idxPrev].Pos, pathTry[iSteps - 1].Pos, NULL)) > 0 && iNewSteps + idxPrev <= iSteps)
				{
					POINT posCur = pathTry[iSteps - 1].Pos;

					// more mem ?
					iSteps = idxPrev + iNewSteps;
					if (iSteps > iReservedSteps)
					{
						iReservedSteps += 1000;
						pathTry = (MPathData *) realloc (pathTry, sizeof (MPathData) * iReservedSteps);
					}

					// insert data
					Interpolate (pathTry[idxPrev].Pos, posCur, &pathTry [idxPrev]);
					pathTry[idxPrev].posZ = -1;

					// no further replacements
					break;
				}
			}
		}
	
		iStepsCurve [dirPath] = iSteps;
		pathCurve [dirPath] = pathTry;
	}
				
	// select the best buffer
	iSteps = 30000;
	MPathData *pathMin = NULL;

	for (dirPath = 0; dirPath < 2; dirPath++)
	{
		if (iStepsCurve[dirPath] < iSteps)
		{
			iSteps = iStepsCurve[dirPath];
			pathMin = pathCurve[dirPath];
		}
	}
	
	if (!pathMin)
	{
		// free all da mem
		for (dirPath = 0; dirPath < 2; dirPath++)
		free (pathCurve [dirPath]);
		return 0;
	}
		
	// copy
	pathStep = new MPathData [iSteps];
	memcpy (pathStep, pathMin, sizeof (MPathData) * iSteps);

	// free all da mem
	for (dirPath = 0; dirPath < 2; dirPath++)
		free (pathCurve [dirPath]);

	return iSteps;
}
